"""Python to HTML.

Example
-------
>>> from htmlclasses import E, to_string
>>> class html(E):
...     class body(E):
...         class p(E):
...             TEXT = 'foo'
...         class p(E):
...             TEXT = 'bar'
...             some_attr = 'note the hyphen in the actual HTML'
...         class p(E):
...             TEXT = 'baz'
...             class_ = 'note trailing underscore'
...
>>> print(to_string(html(), indent='    '))
<!DOCTYPE html>
<html>
    <body>
        <p>foo</p>
        <p some-attr="note the hyphen in the actual HTML">bar</p>
        <p class="note trailing underscore">baz</p>
    </body>
</html>
"""

from .htmlclasses import E  # noqa: F401
from .serialize import to_string  # noqa: F401

__all__ = ('E', 'to_string')

__version__ = '0.3.0'



import inspect
import types

TREES_AND_LEAVES = '_trees_and_leaves'
ELEMENT_ATTRIBUTES = '_element_attributes'

OWNED_ELEMENT_INSTANCES = '_owned_element_instances'
_TEXT_ATTRIBUTE_NAME = 'TEXT'


class _Element:
    pass


def _is_elem_attribute(name):
    return not name.startswith('_') and name != _TEXT_ATTRIBUTE_NAME


def _to_elem_attr_name(name):
    return name.rstrip('_').replace('_', '-')


def _is_elem_class(value):
    return inspect.isclass(value) and issubclass(value, _Element)


def _can_be_converted_to_element_class(name, value):
    # Save a few key strokes by not having to type `class foo(E):`
    # and instead just write `class foo:`.
    #
    # Attributes that meet the requirements will be
    # converted to subclasses of _Element.
    return (
            inspect.isclass(value)
            and not issubclass(value, _Element)
            and not name.startswith('_')
            )


def _create_element_class(name, value):
    """Covert plain `class foo:` to sublcass of `_Element`"""

    def populate_namespace(ns):
        for k, v in value.__dict__.items():
            ns[k] = v

    new = types.new_class(
            name,
            (_Element,),
            kwds=dict(metaclass=_Meta),
            exec_body=populate_namespace,
            )

    return new


class _DictForCollectingElements(dict):

    def __init__(self):
        """Collect some items on the fly.

        Python class cannot have 2 different attributes
        under the same name.
        But given that we want to use Python classes to construct
        HTML elements, we have to circumvent this constraint.
        """
        super().__init__()
        self[TREES_AND_LEAVES] = []
        self[ELEMENT_ATTRIBUTES] = {}

    def __setitem__(self, name, value):
        if _is_elem_class(value) or name == _TEXT_ATTRIBUTE_NAME:
            self[TREES_AND_LEAVES].append(value)
        elif _can_be_converted_to_element_class(name, value):
            self[name] = _create_element_class(name, value)
        elif _is_elem_attribute(name):
            self[ELEMENT_ATTRIBUTES][_to_elem_attr_name(name)] = value
        else:
            super().__setitem__(name, value)


class _Meta(type):

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        d = _DictForCollectingElements()
        children_of_bases = sum(
                [getattr(base, TREES_AND_LEAVES, []) for base in bases],
                [],
                )
        d[TREES_AND_LEAVES].extend(children_of_bases)

        base_element_attributes = {
                k: v
                for base in bases
                for k, v in getattr(base, ELEMENT_ATTRIBUTES, {}).items()
        }
        d[ELEMENT_ATTRIBUTES] = base_element_attributes

        return d

    def __getattr__(cls, name):
        def factory(text, **attributes):
            namespace = _DictForCollectingElements()
            namespace[_TEXT_ATTRIBUTE_NAME] = text
            for k, v in attributes.items():
                namespace[k] = v
            return type(name, (E,), namespace)
        return factory


class E(_Element, metaclass=_Meta):
    """Subclass this element to create an HTML tree.

    Example:

        class html(E):
            class head(E):
                class title(E):
                    TEXT = 'This is a title'

                class meta(E):
                    charset = 'UTF-8'

                class meta(E):
                    # note the duplicated class names - that's fine
                    name = 'description'
                    content = 'usage example'

            class body(E):
                ...
    """

    def __init__(self, *subelements):
        self._subelements = subelements



from typing import NamedTuple, Type, Union
import html

from .htmlclasses import E


def to_string(
        element: Type[E],
        *,
        indent: str = '',
        html_doctype: bool = True,
        ) -> str:
    """Serialize an E instance.

    Parameters
    ----------
    element: n/c.
    indent: If it's given the code will be indented accordingly.
    html_doctype: Whether to prepend the DOCTYPE html declaration.

    Returns
    -------
    Valid HTML string.
    """

    lines = _lines(element, doctype=html_doctype)
    if indent:
        return '\n'.join(
            indent * line.indent_level + line.text
            for line in lines
        )
    else:
        return ''.join(line.text for line in lines)


class IndentedLine(NamedTuple):

    text: str
    indent_level: int


def _lines(
        element: Union[Type[E], E, str],
        *,
        indent_level: int = 0,
        doctype: bool = False,
):
    # For the purpose of pretty formatted HTML it's convenient
    # for me to think of it in terms of collections of lines
    # with indent levels.
    if doctype:
        yield IndentedLine('<!DOCTYPE html>', indent_level)

    if isinstance(element, str):
        for line in element.splitlines():
            yield IndentedLine(html.escape(line), indent_level)
        return

    if isinstance(element, E):
        for e in element._subelements:
            for line in _lines(e, indent_level=indent_level):
                yield line
        return

    tag_name = element.__name__

    is_leaf = not element._trees_and_leaves
    tag_opening = _build_tag_opening(
            tag_name, is_leaf, element._element_attributes)
    tag_closing = _build_tag_closing(tag_name, is_leaf)

    if tag_name.lower() == 'pre':
        yield IndentedLine(
                _handle_pre(element, tag_opening, tag_closing),
                indent_level,
                )
        return

    yield IndentedLine(tag_opening, indent_level)
    for tree_or_leaf in element._trees_and_leaves:
        for indented_line in _lines(
                tree_or_leaf, indent_level=indent_level + 1):
            yield indented_line
    if tag_closing:
        yield IndentedLine(tag_closing, indent_level)


def _build_tag_opening(tag_name, is_leaf, attributes):
    attrs = ' '.join(
            f'{k}="{v}"'
            for k, v in attributes.items()
            )

    prefix = '<' + tag_name
    middle = ' ' + attrs if attrs else ''
    if is_leaf:
        suffix = '/>'
    else:
        suffix = '>'
    return prefix + middle + suffix


def _build_tag_closing(tag_name, is_leaf):
    if is_leaf:
        return ''
    else:
        return f'</{tag_name}>'


def _handle_pre(element, tag_opening, tag_closing):
    text_list = element._trees_and_leaves

    if not text_list or text_list == ['']:
        return tag_opening + tag_closing
    elif len(text_list) != 1 or not isinstance(text_list[0], str):
        raise NotImplementedError(
                f'Do now know what to do with {element}.'
                + ' Is it even legal HTML?'
        )
    else:
        text, = text_list
        return tag_opening + html.escape(text) + tag_closing



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['htmlclasses']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'htmlclasses',
    'version': '0.3.0',
    'description': 'Python in, HTML out.',
    'long_description': '# htmlclasses\n\nPython in, HTML out.\n\nThere are templating engines making it possible to write code\nin HTML template files. However, I would very much prefer\nto be able to write Python that gets converted to HTML \nrather than write Python-like mini language engulfed in HTML. \n\n## Version\n\n0.3.0\n\n## Goals\n\nGenerating valid HTML from pure Python code.\n\n## Non-goals\n\nFeatures geared toward JavaScript.\n\n1. I find using 2 intertwined languages too cumbersome.\n2. JavaScript is heavily overused and misused.\n   I don\'t want to add to the problem.\n\n\n## Installation\n\n`pip install htmlclasses`\n\n## Developing\n\nThis project is managed with poetry: https://github.com/python-poetry/poetry\n\n1. `git clone git@github.com:uigctaw/htmlclasses.git`\n2. `poetry install`\n\n### Running tests\n\n`./check_all.sh`\n\n## Examples\n\nTo convert Python to HTML run:\n\n```python\nfrom htmlclasses import to_string\n\nto_string(html, indent=\'    \')\n```\n\n### Hello World\n\n```python\nfrom htmlclasses import E\n\n\nclass html(E):\n\n    class head:\n        pass\n\n    class body:\n\n        class p:\n\n            TEXT = \'Hello, world!\'\n```\n\n```html\n<!DOCTYPE html>\n<html>\n    <head/>\n    <body>\n        <p>\n            Hello, world!\n        </p>\n    </body>\n</html>\n```\n\n### Repeated Elements\n\n```python\nfrom htmlclasses import E\n\n\nclass html(E):\n\n    class head(E):  # Must sublcass if repeating tags\n\n        class meta:\n            name = \'description\'\n            content = \'Framework usage examples\'\n\n        class meta:  # type: ignore[no-redef]  # noqa: F811\n            name = \'keywords\'\n            content = \'Python, HTML\'\n\n    class body:\n\n        class p:\n\n            TEXT = \'Hello, world!\'\n```\n\n```html\n<!DOCTYPE html>\n<html>\n    <head>\n        <meta name="description" content="Framework usage examples"/>\n        <meta name="keywords" content="Python, HTML"/>\n    </head>\n    <body>\n        <p>\n            Hello, world!\n        </p>\n    </body>\n</html>\n```\n\n## Alternatives\n\nhttps://pypi.org/project/html\n',
    'author': 'uigctaw',
    'author_email': 'uigctaw@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
